/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Regular expression scan routines.
 *
 * One of the following should be defined, depending on which routines your
 * version of Unix supports.  Many Unixes support more than one.  Prefer the
 * version listed first below:
 *
 *  RE_POSIX - Compile with regcomp(), search with regexec().
 *	This exists on Linux, FreeBSD, BSDI and AIX, and is the POSIX 1003.2
 *      standard (section 2.8).  If it recognizes REG_ICASE flag, it is this,
 *      not RE_BELL.  There is a freeware implementation of this by Henry
 *      Spencer, as well as a Gnu version.  (Actually, at least some versions
 *      of the Henry Spencer version seem incompatible with this.)
 *
 *  RE_MATCH - Compile expressions with re_compile(), search with re_match().
 *	This exists on the NeXT.
 *
 *  RE_GEN - Compile with compile() and search with step().
 *	Always case sensitive.  Looks like just a cleaner interface to RE_BSD.
 *	Exists on Solaris and IRIX 5.3, at least.  Needs -lgen flag at compile
 *      time.
 *
 *  RE_GENSGI - Compile expressions with regcmp(), search with regex();
 *      Always case-sensitive.  Needs -lgen at compile time.  Exists on IRIX
 *      Solaris and AIX.
 *
 *  RE_BSD - Compile with compile() and search with step().
 *	Always case sensitive.  Amazingly complicated interface requiring lots
 *	of macro definitions.  This exists on SunOS, IRIX, AIX and various
 *	older bsd unixes.
 *
 *  RE_BELL - Compile with regcomp(), search with regexec().
 *	Always case sensitive.  Originally from Bell V8.  Also in 4.3bsd-Tahoe
 *	and BSDI.
 *
 *  RE_EXEC - Compile expressions with re_comp(), search with re_exec()
 *	Always case-sensitive.  Exists on SunOS, AIX, NeXT and IRIX.
 *
 *  RE_RECMP - Use recmp() call to compile and match
 *	Always case-sensitive.  Pattern recompiled on each call.  Available
 *	on Next.
 *
 * I think I can fake case insensitivity for systems that don't have it by
 * converting the pattern and the text to all lower case before processing.
 * But I haven't done this yet.
 *
 * The following exists, and maybe should be supported, but I haven't got
 * around to it yet.
 *
 *  RE_GNU - Compile with re_compile_pattern(), search with re_search().
 *	The native interface in the GNU regex package.  This same package
 *      provides an RE_POSIX style interface and an RE_EXEC style interface.
 */

#include "backtalk.h"

#include <ctype.h>

#include "sysdict.h"
#include "stack.h"
#include "search.h"

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifdef RE_POSIX
#include <regex.h>
#define DCL	static regex_t preg; int err;
#define FREE	regfree(&preg)
#define CPLFLAG REG_EXTENDED|REG_NOSUB|(bdict(VAR_IGNORE_CASE)?REG_ICASE:0)
#define CPL(p)	if (err= regcomp(&preg, p, CPLFLAG)) \
		{ char errmsg[256]; regerror(err,&preg,errmsg,256); \
		die("Regular expression error: %s",errmsg); }
#define TST(t)	(!regexec(&preg, t,0,(regmatch_t *)NULL,0))
#endif /* RE_POSIX */


#if defined(RE_BSD) || defined(RE_GEN)
void re_die(int v)
{
char *msg;

	switch (v)
	{
	case 11: msg= "Range endpoint too large"; break;
	case 16: msg= "Bad number"; break;
	case 25: msg= "``\\ digit'' out of range"; break;
	case 36: msg= "Illegal or missing delimiter"; break;
	case 41: msg= "No remembered search string"; break;
	case 42: msg= "\\( \\) imbalance"; break;
	case 43: msg= "Too many \\("; break;
	case 44: msg= "More than 2 numbers given in \\{ \\}"; break; /*{*/
	case 45: msg= "} expected after \\"; break;
	case 46: msg= "First number exceeds second in \\{ \\}"; break;
	case 49: msg= "[] imbalance"; break;
	case 50: msg= "Regular expression too long"; break;
	default: msg= "Unknown error"; break;
	}
	die("Regular expression error: %s",msg);
}
#endif /* RE_BSD || RE_GEN */

#ifdef RE_BSD
#define INIT	register char *sp= instring;
#define GETC()	(*(sp++))
#define PEEKC() (*sp)
#define UNGETC(c) (--sp)
#define RETURN(p) return;
#define ERROR(v) re_die(v)
static getrnge();	/* Suppress common error in regexp.h */
#include <regexp.h>
#define DCL static char rebuf[1024];
#define FREE
#define CPL(p)	compile(p, rebuf, rebuf+1024, '\0')
#define TST(t)	step(t,rebuf)
#endif /* RE_BSD */


#ifdef RE_GEN
#include <regexpr.h>
#define DCL	static char *rexp;
#define FREE	free(rexp);
#define CPL(p)  if ((rexp= compile(p, NULL, NULL)) == NULL) re_die(regerrno)
#define TST(t)	step(t,rexp)
#define LOC1	loc1
#define LOC2	loc2
#endif /* RE_GEN */


#ifdef RE_BELL
#include <regexp.h>
void regerror(const char *msg)
{
	die("Regular expression error: %s",msg);
}
#define DCL	static struct regexp *rexp;
#define FREE	free(rexp)
#define CPL(p)	rexp= regcomp(p)
#define TST(t)	regexec(rexp,t)
#endif /* RE_BELL */


#ifdef RE_MATCH
#include <regex.h>
#define DCL	static struct regex *rexp;
#define FREE	free(rexp)
#define CPL(p)	if ((rexp= re_compile(p,bdict(VAR_IGNORE_CASE))) == NULL) \
		die("Regular expression error")
#define TST(t)	re_match(t,rexp)
#endif /* RE_MATCH */


#ifdef RE_EXEC
char *re_comp();
#define DCL	char *err;
#define FREE
#define CPL(p)	if ((err= re_comp(p)) != NULL) \
		die("Regular expression error: %s",err)
#define TST(t)	re_exec(t)
#endif /* RE_EXEC */


#ifdef RE_GENSGI
#include <libgen.h>
#define DCL	static char *rexp;
#define FREE	free(rexp)
#define CPL(p)	if ((rexp= regcmp(p,(char *)NULL)) == NULL) \
		die("Regular expression error")
#define TST(t)	regex(rexp,t)
#endif /* RE_GENSGI */


/* FUNC_GREP: <string> <pattern> grep <bool>
 * This returns non-zero if the given pattern occurs in the given string.
 * The pattern may be an arbitrary regular expression.  Often we do repeated
 * calls with the same pattern, so we remember the previous pattern and don't
 * recompile if it is the same.  If the pattern is null, the previous pattern
 * is repeated.
 */

char *prev_grep= NULL;

#ifdef RE_RECMP

void func_grep()
{
char *pat= pop_string();
char *txt= pop_string();

	if (pat[0] == '\0')
	{
		if (prev_grep == NULL)
			die ("No previous pattern");
		free(pat);
		pat= prev_grep;
	}
	else
	{
		if (prev_grep != NULL)
		{
			free(prev_grep);
			prev_grep= pat;
		}
	}
	
	push_integer(recmp(pat,txt));
	free(pat);
	free(txt);
}

#else
#ifdef RE_NONE

void func_grep()
{
	die("compiled without a regular expression parser");
}

#else

void func_grep()
{
DCL
char *pat= pop_string();
char *txt= pop_string();

	if (pat[0] == '\0')
	{
		if (prev_grep == NULL)
			die("No previous pattern");
		free(pat);
	}
	else if (prev_grep == NULL || strcmp(pat, prev_grep))
	{
		if (prev_grep != NULL)
		{
			FREE;
			free(prev_grep);
		}
		prev_grep= pat;
		CPL(pat);
	}
	else
		free(pat);

	push_integer(TST(txt));

	free(txt);
}

#endif /* RE_NONE */
#endif /* RE_RECMP */



/* BM_COMPILE - Do compilation for a Boyer-Moore Search.  The pattern is
 * expected to be a malloced string, and should *NOT* be freed by the caller
 * afterwards.
 */

char *prev_search= NULL;
int charjump[256];
int *matchjump= NULL;
int prev_len, prev_icase;

int bm_compile(char *pattern)
{
int *bck;
int i, q, r, k;
/* Fake one-based arrays so I don't have to rewrite algorithm */
#define patn (pattern - 1)
#define back (bck - 1)
#define mjmp (matchjump - 1)

	/* FIRST: Save pattern */
	if (pattern[0] == '\0')
	{
		if (prev_search == NULL) die("no previous search");
		free(pattern);
		pattern= prev_search;
	}
	else
	{
		if (prev_search != NULL) free(prev_search);
		prev_search= pattern;
		prev_len= strlen(pattern);
	}
	prev_icase= bdict(VAR_IGNORE_CASE);

	/* SECOND: Build charjump array */

	for (i= 0; i < 256; i++)
		charjump[i]= prev_len;

	for (k= 1; k <= prev_len; k++)
	{
		if (prev_icase && isascii(patn[k]) && isupper(patn[k]))
			patn[k]= tolower(patn[k]);
		charjump[patn[k]]= prev_len - k;
	}
	
	/* FINALLY: Build matchjump array */

	if (matchjump != NULL) free(matchjump);
	matchjump= (int *)malloc(prev_len * sizeof(int));
	bck= (int *)malloc(prev_len * sizeof(int));

	for (k= 1; k <= prev_len; k++)
		mjmp[k]= 2*prev_len - k;
	
	for (k= prev_len, q= prev_len+1; k > 0; k--, q--)
	{
		back[k]= q;
		while (q <= prev_len && patn[k] != patn[q])
		{
			if (mjmp[q] > prev_len - k)
				mjmp[q]= prev_len - k;
			q= back[q];
		}
	}

	for (k= 1; k <= q; k++)
		if (mjmp[k] > prev_len + q - k)
			mjmp[k]= prev_len + q - k;

	r= back[q];
	while (q <= prev_len)
	{
		if (q <= r)
		{
			if (mjmp[q] > r - q + prev_len)
				mjmp[q]= r - q + prev_len;
			q++;
		}
		else
			r= back[r];
	}

	free(bck);

#undef patn
#undef back
#undef mjmp
}


/* BM_SEARCH - Search the given text for the last compiled pattern.  This is
 * the old reliable Boyer-Moore search.  Return -1 if no match, or the offset
 * of the match in the string otherwise.
 */

int bm_search(char *txt)
{
register char *txtp, *patp;
char *txt_end, *pat_end;
char tch;

	txt_end= txt + strlen(txt) - 1;
	pat_end= prev_search + prev_len - 1;

	txtp= txt + prev_len - 1;
	patp= pat_end;

	while (txtp <= txt_end && patp >= prev_search)
	{
		tch= (prev_icase && isascii(*txtp) && isupper(*txtp)) ?
			tolower(*txtp) : *txtp;

		if (tch == *patp)
		{
			txtp--;
			patp--;
		}
		else
		{
			txtp+= max(charjump[tch], matchjump[patp-prev_search]);
			patp= pat_end;
		}
	}
	return ( (patp < prev_search) ? txtp - txt + 1 : -1 );
}


/* FUNC_SEARCH:   <text> <pattern> search <offset>
 * Search a string for a fixed pattern.  Returns the offset into the text at
 * which the pattern was found.  If the pattern is not found, an offset of
 * -1 is pushed.  If the pattern is null, the previous pattern is reused.
 */

void func_search()
{
char *pat= pop_string();
char *txt= pop_string();

	bm_compile(pat);
	push_integer(bm_search(txt));
	free(txt);
	/* do not free(pat) */
}


/* FUNC_REPLACE:  <text> <pattern> <replacement> replace <newtext>
 * Find all occurances of <pattern> in <text> and replace each one with
 * <replacement> resulting in the new string <newtext>.
 */

void func_replace()
{
char *rep= pop_string();
char *pat= pop_string();
char *txt= pop_string();
char *cur= txt;
size_t outsize= strlen(txt)*2;
size_t outlen= 0;
size_t replen= strlen(rep);
int patlen= strlen(pat);
char *out= (char *)malloc(outsize);
int i;

	bm_compile(pat);
	while((i= bm_search(cur)) != -1)
	{
		/* Enlarge output buffer if necessary */
		if (outlen + i + replen + 1 > outsize)
		{
			outsize*= 2;
			out= (char *)realloc(out, outsize);
		}

		/* Copy over start of string */
		strncpy(out+outlen, cur, i);
		outlen+= i;
		/* Copy over replacement */
		strcpy(out+outlen, rep);
		outlen+= replen;

		/* Advance past end of pattern */
		cur+= i + patlen;
	}

	/* Re-allocate buffer to correct size */
	out= (char *)realloc(out, outlen + strlen(cur) + 1);
	/* Copy over last bit */
	strcpy(out+outlen,cur);

	free(txt);
	/* do not free(pat) */
	free(rep);

	push_string(out,FALSE);
}
