/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* USER DATABASE ACCESS ROUTINES
 *
 * This file and udb_read.h implement macros for looking up a user in a
 * secondary database.  The following routines exist for all user database
 * configurations:
 *
 *   getdbnam(login)	- return passwd structure for given conf login name
 *   getdbuid(uid)	- return passwd structure for given conf uid number
 *
 *   firstdb()		- return passwd structure for "first" db entry
 *   nextdb()		- return passwd structure for "next" db entry
 *
 *   gethomedir(login)	- return home directory for given user.
 *
 * The passwd structures returned by the first four routines will have
 * login, uid, gid, fullname, and directory set.  The gethomedir routine
 * returns only the directory path name, and is faster only for tagfile data
 * structures. Shell and passwd fields are normally not used.  Some versions
 * of these routines cache user info, so you can call the following to flush
 * out cached info:
 *
 *      flushusercache() - forget any cached user database info.
 *
 * The primary user database may be accessed with the following routine:
 *
 *	getpassword(login) - return encrypted password for the given user
 *      checkpassword(login,passwd) - check if a login is valid
 *
 * The following two calls are probably the ones we're going to use more than
 * the firstdb() and nextdb().  In all except the passwd file case, these are
 * faster than the looking up the full user data:
 *
 *      firstusr()	- return login name of first user
 *	nextusr() 	- return login name of next user
 *      seekusr(login)  - pretend nextusr just returned this login
 *
 * If UID_MAP_FILE is defined, we also can look up the login name that goes
 * with a uid by calling
 *
 *	mapuid(int uid)
 */

#include <stdio.h>

#include "config.h"

#include <ctype.h>
#include <pwd.h>

#include "set_uid.h"
#include "udb_read.h"
#include "str.h"
#include "lock.h"

#if defined(UID_MAP_FILE) || defined(UD_DBM) || (defined(UD_TAGFILE) && defined(AUTH_DBM))
#include "hashfile.h"
#endif


#ifdef UD_PASSWD

		  /* ------------------------------------- */
		  /* ----- Real or Fake passwd files ----- */
		  /* ------------------------------------- */

struct passwd *last_pw= NULL;

/* FIRSTUSR - (passwd version) - Return name of first user in user database.
 * We actually get the whole passwd structure, but stash it so we don't
 * have to do so again if the script next does a "userinfo" to request it.
 */

char *firstusr()
{
    if (firstdb() == NULL)
	return NULL;
    
    return last_pw->pw_name;
}


/* NEXTUSR - (passwd version) - Return name of next user in user database.
 * We actually get the whole passwd structure, but stash it so we don't
 * have to do so again if the next call to getdbnam() looks it up.
 */

char *nextusr()
{
    if (nextdb() == NULL)
	return NULL;
    else
	return last_pw->pw_name;
}


/* SEEKUSR - (passwd version) - Read through the passwd file until we read
 * the named user.  Future nextusr() calls will continue from there.
 */

void seekusr(char *login)
{
char *u;

    for (u= firstusr(); u != NULL && strcmp(u,login); u= nextusr())
	;
}


/* FLUSHUSERCACHE - (passwd version) - If we do something like change our
 * full name, we need to be able to forget the previously loaded user
 * info and reload it.  This does that.
 */

void flushusercache()
{
    last_pw= NULL;
}

#ifdef PASSWD_FILE

		/* ------------------------------------------ */
		/* ---------- fake passwd-like file --------- */
		/* ------------------------------------------ */

#ifndef HAVE_FGETPWENT

/* FGETPWENT - Read a passwd-file-like line from the given file descriptor.
 * This is a built-in function under Solaris, but not apparantly much of
 * anyplace else.  THIS IS NOT VERY ROBUST!  It skips lines it doesn't
 * understand.
 */

struct passwd *fgetpwent(FILE *fp)
{
static char bf[BFSZ+1];
static struct passwd pw;
char *p, *b, *e;

    while ((p= fgets(bf, BFSZ, fp)) != NULL)
    {
	/* Skip lines longer than BFSZ */
	if ((p= strchr(bf,'\n')) == NULL)
	{
	    int ch;
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;
	}
	else
	{
	    /* Get login name */
	    b= bf;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.pw_name= b; *e= '\0';

	    /* Get password */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.pw_passwd= b; *e= '\0';

	    /* Get uid */
	    b= e+1;
	    if (!isascii(*b) || !isdigit(*b)) continue;
	    if ((e= strchr(++b,':')) == NULL) continue;
	    pw.pw_uid= atoi(b);

	    /* Get gid */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.pw_gid= atoi(b);
	    
	    /* Get fullname */
	    b= e+1;
	    if ((e= strchr(b,':')) == NULL) continue;
	    pw.pw_gecos= b; *e= '\0';
	    
	    /* Get directory */
	    b= e+1;
	    pw.pw_dir= b;
	    if ((e= strchr(b,':')) != NULL) *e= '\0';

	    return &pw;
	}
    }
    return NULL;

}
#endif /* HAVE_FGETPWENT */


/* GETPWNU - Get a particular name or uid from PASSWD_FILE.  Either the name
 * should be NULL or the uid should be -1, to indicate that the other is the
 * one we are searching for.  This is not a very fast implementation.
 */

struct passwd *getpwnu(char *nam, uid_t uid)
{
FILE *pwfp;
struct passwd *pw;

    if (nam != NULL && last_pw != NULL && !strcmp(nam, last_pw->pw_name))
	return last_pw;

    set_httpd_if_shared();
    if ((pwfp= fopen(PASSWD_FILE,"r")) == NULL)
    {
	set_cfadm_if_shared();
	die("getpwnu cannot open user database "PASSWD_FILE);
    }
    lock_shared(fileno(pwfp),PASSWD_FILE);

    while ((pw= fgetpwent(pwfp)) != NULL)
    {
	if ((uid >= 0 && pw->pw_uid == uid) ||
	    (nam != NULL && !strcmp(nam, pw->pw_name)))
		break;
    }

    fclose(pwfp);
    unlock(PASSWD_FILE);
    set_cfadm_if_shared();
    return last_pw= pw;
}


/* RESETPW - Open or rewind the database file.  This runs without locks,
 * which may occasionally give glitches, but this is never used for
 * anything very critical (like authentication) and we have no way of
 * knowing when we are done with the file, so we'd have to hold the lock
 * till we exit.
 */

FILE *dbfp= NULL;
void resetpw()
{
    if (dbfp == NULL)
    {
	set_httpd_if_shared();
	dbfp= fopen(PASSWD_FILE,"r");
	set_cfadm_if_shared();
	if (dbfp == NULL)
	    die("resetpw cannot open user database "PASSWD_FILE);
    }
    else
	rewind(dbfp);
}

#else /* ! PASSWD_FILE */

		   /* ------------------------------------- */
		   /* ---------- real passwd file --------- */
		   /* ------------------------------------- */

/* GETDBNAM - (real passwd file version) - Find passwd structure for the
 * named user.  Really just a wrapper for getpwnam() that first checks if
 * we already have the info on that user.
 */

struct passwd *getdbnam(char *name)
{
    if (last_pw != NULL && !strcmp(name, last_pw->pw_name))
	return last_pw;
    else
	return (last_pw= getpwnam(name));
}

#endif /* ! PASSWD_FILE */
#else  /* ! UD_PASSWD */
#ifdef UD_DBM

                  /* ------------------------------------ */
                  /* ----------- dbm database ----------- */
                  /* ------------------------------------ */

/* GETDBNAM - Get user data from the DBM_NAM_FILE.  Save info into a
 * standard passwd structure in static storage.  Return NULL on failure.
 *
 * DBM_NAM_FILE entries look like:
 *   <encryptpassswd>:<uid>:<gid>:<fullname>:<directory>
 */

struct passwd *getdbnam(char *nam)
{
static char bf[BFSZ];
static char name[MAX_LOGIN_LEN+1];
static struct passwd pw;
char *e, *b;
int rc;

    /* Fetch the line from the db file */
    set_httpd_if_shared();
    rc= fetchdbm(DBM_NAM_FILE, nam, strlen(nam), bf, BFSZ);
    set_cfadm_if_shared();

    /* Return NULL if not found */
    if (rc) return NULL;

    /* Copy name into static storage */
    strcpy(name,nam);
    pw.pw_name= name;

    /* Get password */
    if ((e= strchr(bf,':')) == NULL) return NULL;
    pw.pw_passwd= bf; *e= '\0';

    /* Get uid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_uid= atoi(b);

    /* Get gid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_gid= atoi(b);
    
    /* Get fullname */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_gecos= b; *e= '\0';
    
    /* Get directory */
    b= e+1;
    pw.pw_dir= b;
    if ((e= strchr(b,':')) != NULL) *e= '\0';

    return &pw;
}

#if defined(ID_GETUID) && defined(UID_MAP_FILE) && defined(DBM_UID_FILE)
/* OK, so we never really need this... */

/* GETDBUID - Get user data from the DBM_UID_FILE.  Save info into a
 * standard passwd structure in static storage.  Return NULL on failure.
 *
 * DBM_UID_FILE entries look like:
 *   <encryptpassswd>:<loginuid>:<gid>:<fullname>:<directory>
 */

struct passwd *getdbuid(uid_t uid)
{
static char bf[BFSZ];
static struct passwd pw;
char *e, *b;
int rc;

    /* Fetch the line from the db file */
    set_httpd_if_shared();
    rc= fetchdbm(DBM_UID_FILE, (char *)&uid, sizeof(uid), bf, BFSZ);
    set_cfadm_if_shared();

    /* Return NULL if not found */
    if (rc) return NULL;

    /* Get password */
    if ((e= strchr(bf,':')) == NULL) return NULL;
    pw.pw_passwd= bf; *e= '\0';

    /* Get name */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_name= b; *e= '\0';

    /* Save uid */
    pw.pw_uid= uid;

    /* Get gid */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_gid= atoi(b);
    
    /* Get fullname */
    b= e+1;
    if ((e= strchr(b,':')) == NULL) return NULL;
    pw.pw_gecos= b; *e= '\0';
    
    /* Get directory */
    b= e+1;
    pw.pw_dir= b;
    if ((e= strchr(b,':')) != NULL) *e= '\0';
    
    return &pw;
}
#endif /* ID_GETUID && UID_MAP_FILE && DBM_UID_FILE */


/* FIRSTUSR - (dbm version) - Return the "first" user in the dbm file.
 */

char *firstusr()
{
static char bf[MAX_LOGIN_LEN+1];

    set_httpd_if_shared();
    if (walkdbm(DBM_NAM_FILE))
    {
	set_cfadm_if_shared();
	die("firstusr cannot open name db "DBM_NAM_FILE" to read");
    }
    set_cfadm_if_shared();
    
    if (firstdbm(bf, MAX_LOGIN_LEN+1))
	return NULL;
    else
	return bf;
}


/* NEXTUSR - (dbm version) - Return the "next" user in the dbm file.
 */

char *nextusr()
{
static char bf[MAX_LOGIN_LEN+1];

    if (nextdbm(bf, MAX_LOGIN_LEN+1))
	return NULL;
    else
	return bf;
}


/* SEEKUSR - (dbm version) - Set the scan to start after the named
 * user.  Future nextusr() calls will continue from there.
 */

void seekusr(char *login)
{
    set_httpd_if_shared();

    if (walkdbm(DBM_NAM_FILE))
    {
	set_cfadm_if_shared();
	die("seekusr cannot open name db "DBM_NAM_FILE" to read");
    }
    set_cfadm_if_shared();
    
    seekdbm(login, strlen(login));
}
#else /* ! UD_DBM */
#ifdef UD_TAGFILE

		/* ------------------------------------------------ */
                /* ---------- tagfiles in user directories -------- */
		/* ------------------------------------------------ */

/* GETHOMEDIR - Return the user's home directory.  Version for tagfiles.
 */

char *gethomedir(char *name)
{
static char dir[BFSZ];

#if USER_DIR_LEVEL == 0
    sprintf(dir, USER_DIR"%s", name);
#elif USER_DIR_LEVEL == 1
    sprintf(dir, USER_DIR"%c/%s", name[0], name);
#else
    sprintf(dir, USER_DIR"%c/%c/%s",
	name[0], name[1] == '\0' ? '_' : name[1], name);
#endif
    return dir;
}


/* GETDBNAM - Get the database entry for the named user.  (This is only for
 * UD_TAGFILE).
 */

struct passwd *getdbnam(char *nam)
{
static struct passwd pw;
static char fn[BFSZ];
char fname[BFSZ+1];
char bf[BFSZ+1], *dir, *var, *val, *end;
FILE *fp;
int ch;

    pw.pw_name= nam;

    /* Construct the directory name */
    if (nam == NULL || (dir= gethomedir(nam)) == NULL)
    	return NULL;
    pw.pw_dir= dir;

    /* Open tag file */
    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"r")) == NULL)
	    return NULL;
    lock_shared(fileno(fp),fname);
    
    /* Read through tag file */
    while (fgets(bf,BFSZ,fp) != NULL)
    {
	/* Parse apart the line */
	var= bf+ (bf[0] == '*');
	if ((val= strchr(var, '=')) == NULL)
	{
	    fclose(fp);
	    unlock(fname);
	    return NULL;
	}
	*(val++)= '\0';
	end= val + strlen(val) - 1;
	if (*end == '\n')
	    *(end--)= '\0';
	else
	    /* skip over rest of over-long line */
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;

	if (!strcmp(var, "uid"))
	{
	    if (!isascii(*val) || !isdigit(*val))
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    pw.pw_uid= atoi(val);
	}
	else if (!strcmp(var, "gid"))
	{
	    if (!isascii(*val) || !isdigit(*val))
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    pw.pw_gid= atoi(val);
	}
	else if (!strcmp(var, "fullname"))
	{
	    /* We recognize both curly and round braces so both
	     * old and new file format works
	     */
	    if (*val != '(' && *val != '{')
	    {
		fclose(fp);
		unlock(fname);
		return NULL;
	    }
	    if (*end == ')' || *end == '}') *end= '\0';
	    unesctag(val+1,fn,0);
	    pw.pw_gecos= fn;
	}
    }

    fclose(fp);
    unlock(fname);
    return &pw;
}

#ifdef AUTH_FILE

FILE *htpfp= NULL;

/* FIRSTUSR() - (tagfile/file version) - Get the first name out of the
 * htpasswd file.
 */

char *firstusr()
{
    if (htpfp == NULL)
    {
	set_httpd();
	htpfp= fopen(AUTH_FILE,"r");
	set_cfadm();
	if (htpfp == NULL)
		die("firstusr cannot open pwfile "AUTH_FILE" to read");
    }
    else
	rewind(htpfp);
    return nextusr();
}

/* NEXTUSR() - (tagfile/file version) - Get the name of the next user in the
 * htpasswd file.
 */

char *nextusr()
{
static char bf[BFSZ+1];

    if (htpfp == NULL ||
	fgets(bf,BFSZ,htpfp) == NULL)
	    return NULL;
    
    *firstin(bf,":\n")= '\0';
    return bf;
}

/* SEEKUSR() - (tagfile/file version) - Search to the location the hard way.
 */

void seekusr(char *login)
{
char *u;

	for (u= firstusr(); u != NULL && strcmp(u,login); u= nextusr())
		;
}

#endif /* AUTH_FILE */

#ifdef AUTH_DBM

/* FIRSTUSR() - (tagfile/dbm version) - Get the first name out of the htdbm
 * file.
 */

char *firstusr()
{
static char bf[MAX_LOGIN_LEN+1];

    set_httpd();
    if (walkdbm(AUTH_DBM))
    {
	set_cfadm();
	die("firstusr cannot open db "AUTH_DBM" to read");
    }
    set_cfadm();

    if (firstdbm(bf, MAX_LOGIN_LEN+1))
	return NULL;
    else
	return bf;
}

/* NEXTUSR() - (tagfile/dbm version) - Get the name of the next user in the
 * htdbm file.
 */

char *nextusr()
{
static char bf[MAX_LOGIN_LEN+1];

    if (nextdbm(bf, MAX_LOGIN_LEN+1))
	return NULL;
    else
	return bf;
}


/* SEEKUSR() - (tagfile/dbm version) - Seek to the named position in the
 * htdbm file.
 */

void seekusr(char *login)
{
    set_httpd();
    if (walkdbm(AUTH_DBM))
    {
	set_cfadm();
	die("seekusr cannot open db "AUTH_DBM" to read");
    }
    set_cfadm();

    seekdbm(login, strlen(login));
}

#endif

#endif /* UD_TAGFILE */
#endif /* ! UD_DBM */
#endif /* ! UD_PASSWD */


#if defined(ID_GETUID) && defined(UID_MAP_FILE)

/* MAPUID - Given a uid number, use the UID_MAP_FILE dbm to convert it to a
 * user name.  Return NULL on failure.  The UID_MAP_FILE uses the 4-byte
 * uid number as a key, and the content is just the user name.
 */

char *mapuid(uid_t uid)
{
static char name[MAX_LOGIN_LEN+1];
int rc;

    switch (fetchdbm(UID_MAP_FILE, (char *)&uid, sizeof(uid),
    		     bf, MAX_LOGIN_LEN+1))
    {
    case 1:  die("could not open "UID_MAP_FILE" to read");
    case 0:  return name;
    default: return NULL;
    }
}

#endif /* ID_GETUID && UID_MAP_FILE */


/* GETPASSWORD - Get the encrypted password for a user.  Returns NULL if there
 * is no such user. */

#ifdef SHARE_UPW

char *getpassword(char *nam)	/* Password is in the secondary database */
{
struct passwd *pw;
static char pass[40];

    /* Structure returned by getdbnam() has password for shared files */
    if ((pw= getdbnam(nam)) == NULL)
	return NULL;
    strcpy(pass,pw->pw_passwd);
    return pass;
}

#else /* !SHARE_UPW */
#ifdef AUTH_FILE

char *getpassword(char *nam)	/* Password is in flat htpasswd file */
{
char bf[BFSZ+1];
static char pass[40];
char *a,*b;
int n;
FILE *fp;

    set_httpd();
    fp= fopen(AUTH_FILE,"r");
    set_cfadm();
    if (fp == NULL) die("getpassword cannot open file "AUTH_FILE" to read");

    n= strlen(nam);

    while (fgets(bf,BFSZ,fp) != NULL)
    {
	if (bf[n] == ':' && !strncmp(bf,nam,n))
	{
	    a= pass; b= bf+n+1;
	    while (*b != ':' && *b != '\n' && *b != '\0')
		 *(a++)= *(b++);
	    *a= '\0';
	    fclose(fp);
	    return pass;
	}
    }
    fclose(fp);
    return NULL;
}

#endif /* AUTH_FILE */
#ifdef AUTH_DBM

char *getpassword(char *nam)	/* Password is in apache-style dbm */
{
static char pass[40];
int rc;

    set_httpd();
    rc= fetchdbm(AUTH_DBM, nam, strlen(nam), pass, 40);
    set_cfadm();

    switch (rc)
    {
    case 0:  return pass;
    case 1:  die("getpassword cannot open db "AUTH_DBM" to read");
    default: return NULL;
    }
}

#endif /* AUTH_DBM */
#ifdef AUTH_UNIX

char *getpassword(char *name)
{
    /* Unix passwords are most likely in a shadow file somewhere where we
     * can't get them.  For now, don't even bother trying.
     */
    return "";
}

#endif /* AUTH_UNIX */
#endif /* !SHARE_UPW */



/* GETHOMEDIR - Return the user's home directory.  Version for everything
 * except tagfiles.
 */

#ifndef UD_TAGFILE
char *gethomedir(char *name)
{
struct passwd *pw;

    if ((pw= getdbnam(name)) == NULL)
    	return NULL;
    else
    	return pw->pw_dir;
}
#endif


#ifdef ID_SESSION

/* CHECKPASSWORD - Given a login and password, return TRUE if they are
 * are correct.
 */

#if defined(SHARE_UPW) || !defined(AUTH_UNIX)

int checkpassword(char *login, char *passwd)	/* Our own database */
{
char *encpass;

    if ((encpass= getpassword(login)) != NULL)
	return !strcmp(encpass,crypt(passwd,encpass));
    else
	return 0;
}

#else
#ifdef SHADOW_NONE

int checkpassword(char *login, char *passwd)	/* Readable Unix database */
{
struct passwd *pw;

    if ((pw= getpwnam(login)) != NULL)
	return !strcmp(pw->pw_passwd,crypt(passwd,pw->pw_passwd));
    else
	return 0;
}

#else

int checkpassword(char *login, char *passwd)	/* Unreadable Unix database */
{
pid_t child_pid, wait_pid;
waittype wrc;
int pip[2];

    if (pipe(pip))
	return 0;

    switch(child_pid = fork())
    {
    case 0: /* Child - exec the command */

	close(pip[1]);

	/* Attach pip[0] to child's stdin */
	if (pip[0] != 0)
	{
	    dup2(pip[0],0);
	    close(pip[0]);
	}

	execl(PWAUTH_PATH, (char *)NULL);
	/* Exec failed */ exit(-1);

    case -1: /* Parent - Fork failed */

	close(pip[0]);
	close(pip[1]);

	return 0;

    default: /* Parent - Fork succeeded */

	close(pip[0]);

	/* Send the data down the pipe */
	write(pip[1], login, strlen(login));
	write(pip[1], "\n", 1);
	write(pip[1], passwd, strlen(passwd));
	write(pip[1], "\n", 1);

	close(pip[1]);

	/* Wait for child process to exit */
	do {
	    if ((wait_pid = wait(&wrc)) == -1)
		return 0;
	} while (wait_pid != child_pid);

	/* If child was killed, deny authentication */
	if (!WIFEXITED(wrc))
	    return 0;

	/* Child exited cleanly, return status */
	
	return !WEXITSTATUS(wrc);
    }
}

#endif /* SHADOW_NONE */
#endif /* SHARE_UPW || !AUTH_UNIX */

#endif /* ID_SESSION */
